# -*- coding: utf-8 -*-
from contextlib import contextmanager
from datetime import datetime
from functools import partial

import odoo
from odoo.exceptions import AccessDenied
from odoo.tools import config, DEFAULT_SERVER_DATETIME_FORMAT
from odoo.tools.misc import mute_logger

from odoo.addons.base.models.ir_cron import ir_cron

from odoo.service.server import WorkerCron

from .const import LOCAL
from .demo import DEMO, DEMO_PREFIX
from .compat import http, request
from .util import has_role
from . import metabase

from odoo.service import db
bind = wraps = lambda x: x


def _get_port():
    port = config.get('http_port')
    if getattr(odoo, 'evented', False):
        # longpolling process overwrite the config...
        port -= 3
    return port


if DEMO:
    def _db_filter(dbs, httprequest=None):
        return [db for db in (dbs or []) if db.startswith(DEMO_PREFIX)]
else:
    def _db_filter(dbs, httprequest=None):
        r = httprequest or request.httprequest
        if r.headers.get('X-Odoo-Database-Name'):
            return [r.headers['X-Odoo-Database-Name']]
        db = metabase.get_by_host(r.host)
        return [db.name] if db else []

def _db_list(force=False, httprequest=None):
    return _db_filter(None, httprequest)


if hasattr(http, 'Request'):
    # >= saas-15.3
    @contextmanager
    def patch_current_request(req):
        global request
        orig_request, request = request, req
        try:
            yield
        finally:
            request = orig_request

    # http.Request.__init__ will call _db_filter, which needs request to be populated
    # in order to access the httprequest. To break the dependency loop, we temporarily set
    # request to the request being built
    http_Request_init = http.Request.__init__
    def new_http_Request_init(self, httprequest):
        self.httprequest = httprequest
        with patch_current_request(self):
            return http_Request_init(self, httprequest)
    http.Request.__init__ = new_http_Request_init

    # Signatures changed
    _db_filter_pre153 = _db_filter
    def _db_filter(dbs, host=None):
        return _db_filter_pre153(dbs, httprequest=request.httprequest)

    _db_list_pre153 = _db_list
    def _db_list(force=False, host=None):
        return _db_list_pre153(force=force, httprequest=request.httprequest)

def exp_list(document=False):
    raise AccessDenied()

def cron_db_list(self):
    if has_role('slave'):
        return []
    port = _get_port()

    dbname, inround, total = metabase._next_cron(port)

    if not dbname:
        return []

    dbs = [dbname]

    # db_index is increased before getting the database name from the list (thus the dummy entry).
    # sleep is done when index is 0

    if inround == 1:
        self.db_index = -1
    else:
        self.db_index = 0
        dbs = ['dummy'] + dbs

    return dbs


@mute_logger(metabase._logger.name)
def _acquire_job(db_name):
    JUMP = 5
    db = metabase.get(db_name)
    to_process = True
    if db:
        to_process = not db.cron_time or db.cron_time < datetime.now()
        if to_process:
            metabase.update(db, cron_round=db.cron_round + JUMP, cron_time=None)

    try:
        if to_process:
            return _orig_acquire_job(db_name)
    finally:
        if db:
            values = {
                'cron_round': db.cron_round + 1,
            }
            db = metabase.get(db_name)
            if not db.cron_time:
                # The next cron time was not set in meta:
                # - for trials, this means no job was processed
                # - for others, this simply means _process_job is not overriden to set the next cron time in meta
                cron_time = datetime.now()
                values['cron_time'] = cron_time.strftime(DEFAULT_SERVER_DATETIME_FORMAT)
            metabase.update(db, **values)


@mute_logger(metabase._logger.name)
def _process_jobs(db_name):
    JUMP = 5
    db = metabase.get(db_name)
    to_process = True
    if db:
        to_process = not db.cron_time or db.cron_time < datetime.now()
        if to_process:
            metabase.update(db, cron_round=db.cron_round + JUMP, cron_time=None)

    try:
        if to_process:
            return _orig_process_jobs(db_name)
    finally:
        if db:
            values = {
                'cron_round': db.cron_round + 1,
            }
            db = metabase.get(db_name)
            if not db.cron_time:
                # The next cron time was not set in meta:
                # - for trials, this means no job was processed
                # - for others, this simply means _process_job is not overriden to set the next cron time in meta
                cron_time = datetime.now()
                values['cron_time'] = cron_time.strftime(DEFAULT_SERVER_DATETIME_FORMAT)
            metabase.update(db, **values)


# MONKEY PATCH !!!
wrapper = wraps(exp_list)
wrapper._wrapped = bind(db.exp_list)
db.exp_list = wrapper
db.exp_db_exist = wrapper

if not LOCAL and not has_role('runbot'):
    http.db_filter = _db_filter
    http.db_list = _db_list

    # cron patches
    WorkerCron._db_list = cron_db_list
    try:
        ir_cron._acquire_job
    except AttributeError:
        # >= SaaS-14.2
        _orig_process_jobs = ir_cron._process_jobs
        ir_cron._process_jobs = partial(_process_jobs)
    else:
        _orig_acquire_job = ir_cron._acquire_job
        # HACK: use partial to avoid having an unbound method
        ir_cron._acquire_job = partial(_acquire_job)

if bool(config["db_name"] and (config['test_enable'] or config['test_file'])):
    from .tests import util as tests_util
    tests_util._patch_dbfilter_for_tests()
